/* MDXCanvas syntax highlighting CSS with !important to override Canvas styles */
code.highlight { padding: 2px 4px !important; } /* Add padding to inline code blocks to match Canvas default, it is reset by our manual styling otherwise */
div.highlight + div.highlight { margin-top: 1rem !important; } /* Add spacing between consecutive code blocks */
div[data-lines][data-source] + div[data-lines][data-source] { margin-top: 1rem !important; } /* Add spacing between consecutive code blocks from include tags */
.highlight .filename { display: block !important; background: #e0e0e0 !important; padding: 4px 8px !important; border-radius: 6px 6px 0 0 !important; font-family: monospace !important; font-size: .85rem !important; } /* Add a filename header above code blocks if needed */
.highlight .filename + pre { border-radius: 0 0 6px 6px !important; } /* Add a filename header above code blocks if needed */
